@@ -1,3 +1,4 @@
+{-# Language BlockArguments #-}
 {-|
 Module      : Hookup
 Description : Network connections generalized over TLS and SOCKS
@@ -30,6 +31,7 @@ module Hookup
   connect,
   connectWithSocket,
   close,
+  upgradeTls,
 
   -- * Reading and writing data
   recv,
@@ -41,6 +43,7 @@ module Hookup
   ConnectionParams(..),
   SocksParams(..),
   TlsParams(..),
+  TlsVerify(..),
   PEM.PemPasswordSupply(..),
   defaultTlsParams,
 
@@ -60,6 +63,8 @@ module Hookup
   , getPeerPubkeyFingerprintSha512
   ) where
 
+import           Control.Concurrent.Async
+import           Control.Concurrent.STM
 import           Control.Concurrent
 import           Control.Exception
 import           Control.Monad
@@ -69,6 +74,9 @@ import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as B8
 import           Data.Foldable
 import           Data.List (intercalate, partition)
+import           Data.Maybe (fromMaybe, mapMaybe)
+import           Foreign.C.String (withCStringLen)
+import           Foreign.Ptr (nullPtr)
 import           Network.Socket (AddrInfo, HostName, PortNumber, SockAddr, Socket, Family)
 import qualified Network.Socket as Socket
 import qualified Network.Socket.ByteString as SocketB
@@ -83,10 +91,10 @@ import qualified OpenSSL.EVP.Digest as Digest
 import           Data.Attoparsec.ByteString (Parser)
 import qualified Data.Attoparsec.ByteString as Parser
 
-import           Hookup.OpenSSL (installVerification, getPubKeyDer)
+import           Hookup.Concurrent (concurrentAttempts)
+import           Hookup.OpenSSL
 import           Hookup.Socks5
 
-
 -- | Parameters for 'connect'.
 --
 -- Common defaults for fields: 'defaultFamily', 'defaultTlsParams'
@@ -113,24 +121,30 @@ data SocksParams = SocksParams
   , spPort :: PortNumber -- ^ SOCKS server port
   }
 
-
 -- | TLS connection parameters. These parameters are passed to
 -- OpenSSL when making a secure connection.
 data TlsParams = TlsParams
   { tpClientCertificate  :: Maybe FilePath -- ^ Path to client certificate
   , tpClientPrivateKey   :: Maybe FilePath -- ^ Path to client private key
-  , tpClientPrivateKeyPassword :: PEM.PemPasswordSupply -- ^ Private key decryption password
+  , tpClientPrivateKeyPassword :: Maybe ByteString -- ^ Private key decryption password
   , tpServerCertificate  :: Maybe FilePath -- ^ Path to CA certificate bundle
   , tpCipherSuite        :: String -- ^ OpenSSL cipher suite name (e.g. @\"HIGH\"@)
-  , tpInsecure           :: Bool -- ^ Disables certificate checking when 'True'
+  , tpCipherSuiteTls13   :: Maybe String -- ^ OpenSSL cipher suites for TLS 1.3
+  , tpVerify             :: TlsVerify -- ^ Hostname to use when checking certificate validity
   }
 
+data TlsVerify
+  = VerifyDefault -- ^ Use the connection hostname to verify
+  | VerifyNone -- ^ No verification
+  | VerifyHostname String -- ^ Use the given hostname to verify
+  deriving Show
+
 -- | Type for errors that can be thrown by this package.
 data ConnectionFailure
   -- | Failure during 'getAddrInfo' resolving remote host
   = HostnameResolutionFailure HostName String
   -- | Failure during 'connect' to remote host
-  | ConnectionFailure [IOError]
+  | ConnectionFailure [ConnectError]
   -- | Failure during 'recvLine'
   | LineTooLong
   -- | Incomplete line during 'recvLine'
@@ -174,16 +188,23 @@ instance Exception ConnectionFailure where
       AddrNotSupported  -> "address type not supported"
       CommandReply n    -> "unknown reply " ++ show n
 
+data ConnectError = ConnectError SockAddr IOError
+  deriving Show
+
+instance Exception ConnectError where
+  displayException (ConnectError addr e) = show addr ++ ": " ++ displayException e
+
 -- | Default values for TLS that use no client certificates, use
 -- system CA root, @\"HIGH\"@ cipher suite, and which validate hostnames.
 defaultTlsParams :: TlsParams
 defaultTlsParams = TlsParams
   { tpClientCertificate  = Nothing
   , tpClientPrivateKey   = Nothing
-  , tpClientPrivateKeyPassword = PEM.PwNone
+  , tpClientPrivateKeyPassword = Nothing
   , tpServerCertificate  = Nothing -- use system provided CAs
   , tpCipherSuite        = "HIGH"
-  , tpInsecure           = False
+  , tpCipherSuiteTls13   = Nothing
+  , tpVerify             = VerifyDefault
   }
 
 ------------------------------------------------------------------------
@@ -259,12 +280,15 @@ openSocket' h p mbBind =
      let pairs = interleaveAddressFamilies (matchBindAddrs mbSrc dst)
      when (null pairs)
        (throwIO (HostnameResolutionFailure h "No source/destination address family match"))
-     attempt pairs
+     res <- concurrentAttempts connAttemptDelay Socket.close (uncurry connectToAddrInfo <$> pairs)
+     case res of
+       Left es -> throwIO (ConnectionFailure (mapMaybe fromException es))
+       Right s -> pure s
 
 hints :: AddrInfo
 hints = Socket.defaultHints
   { Socket.addrSocketType = Socket.Stream
-  , Socket.addrFlags      = [Socket.AI_ADDRCONFIG, Socket.AI_NUMERICSERV]
+  , Socket.addrFlags      = [Socket.AI_NUMERICSERV]
   }
 
 resolve :: Maybe PortNumber -> HostName -> IO [AddrInfo]
@@ -294,36 +318,6 @@ matchBindAddrs (Just src) dst =
 connAttemptDelay :: Int
 connAttemptDelay = 150 * 1000 -- 150ms
 
-attempt ::
-  [(Maybe SockAddr, AddrInfo)] {- ^ candidate AddrInfos -} ->
-  IO Socket         {- ^ connected socket    -}
-attempt xs =
-  do comm <- newEmptyMVar
-
-     let mkThread i (mbSrc, ai) =
-           forkIOWithUnmask $ \unmask ->
-           unmask $
-           do threadDelay (connAttemptDelay * i)
-              putMVar comm =<< try (connectToAddrInfo mbSrc ai)
-
-     bracket (zipWithM mkThread [0..] xs)
-             (traverse_ killThread)
-             (\_ -> gather (length xs) [] comm)
-
--- Either gather all of the errors possible and throw an exception or
--- return the first successful socket.
-gather ::
-  Int {- ^ potential errors remaining -} ->
-  [IOError] {- ^ errors gathered so far -} ->
-  MVar (Either IOError Socket) ->
-  IO Socket
-gather 0 exs _ = throwIO (ConnectionFailure exs)
-gather n exs comm =
-  do res <- takeMVar comm
-     case res of
-       Right s -> pure s
-       Left ex -> gather (n-1) (ex:exs) comm
-
 -- | Alternate list of addresses between IPv6 and other (IPv4) addresses.
 interleaveAddressFamilies :: [(Maybe SockAddr, AddrInfo)] -> [(Maybe SockAddr, AddrInfo)]
 interleaveAddressFamilies xs = interleave sixes others
@@ -339,10 +333,12 @@ interleaveAddressFamilies xs = interleave sixes others
 -- by the given 'AddrInfo' and return the connected socket.
 connectToAddrInfo :: Maybe SockAddr -> AddrInfo -> IO Socket
 connectToAddrInfo mbSrc info
-  = bracketOnError (socket' info) Socket.close $ \s ->
+  = let addr = Socket.addrAddress info in
+    bracketOnError (socket' info) Socket.close $ \s ->
     do traverse_ (bind' s) mbSrc
-       Socket.connect s (Socket.addrAddress info)
+       Socket.connect s addr
        pure s
+    `catch` (throwIO . ConnectError addr)
 
 -- | A version of 'Socket.bind' that doesn't bother binding on the wildcard
 -- address. The effect of binding on a wildcard address in this library
@@ -404,7 +400,10 @@ networkRecv (SSL  _ s) = SSL.read     s
 -- | A connection to a network service along with its read buffer
 -- used for line-oriented protocols. The connection could be a plain
 -- network connection, SOCKS connected, or TLS.
-data Connection = Connection (MVar ByteString) NetworkHandle
+data Connection =
+  Connection
+  {-# UNPACK #-} !(MVar ByteString)
+  {-# UNPACK #-} !(MVar NetworkHandle)
 
 -- | Open network connection to TCP service specified by
 -- the given parameters.
@@ -418,8 +417,7 @@ connect ::
   IO Connection    {- ^ open connection -}
 connect params =
   do h <- openNetworkHandle params (openSocket params)
-     b <- newMVar B.empty
-     return (Connection b h)
+     Connection <$> newMVar B.empty <*> newMVar h
 
 -- | Create a new 'Connection' using an already connected socket.
 -- This will attempt to start TLS if configured but will ignore
@@ -433,14 +431,13 @@ connectWithSocket ::
   IO Connection    {- ^ open connection  -}
 connectWithSocket params sock =
   do h <- openNetworkHandle params (return sock)
-     b <- newMVar B.empty
-     return (Connection b h)
+     Connection <$> newMVar B.empty <*> newMVar h
 
 -- | Close network connection.
 close ::
   Connection {- ^ open connection -} ->
   IO ()
-close (Connection _ h) = closeNetworkHandle h
+close (Connection _ m) = withMVar m $ \h -> closeNetworkHandle h
 
 -- | Receive the next chunk from the stream. This operation will first
 -- return the buffer if it contains a non-empty chunk. Otherwise it will
@@ -451,11 +448,13 @@ recv ::
   Connection    {- ^ open connection              -} ->
   Int           {- ^ maximum underlying recv size -} ->
   IO ByteString {- ^ next chunk from stream       -}
-recv (Connection buf h) n =
-  do bufChunk <- swapMVar buf B.empty
-     if B.null bufChunk
-       then networkRecv h n
-       else return bufChunk
+recv (Connection bufVar hVar) n =
+  modifyMVar bufVar $ \bufChunk ->
+  do if B.null bufChunk
+       then do h <- readMVar hVar
+               bs <- networkRecv h n
+               return (B.empty, bs)
+       else return (B.empty, bufChunk)
 
 -- | Receive a line from the network connection. Both
 -- @"\\r\\n"@ and @"\\n"@ are recognized.
@@ -472,14 +471,15 @@ recvLine ::
   Connection            {- ^ open connection            -} ->
   Int                   {- ^ maximum line length        -} ->
   IO (Maybe ByteString) {- ^ next line or end-of-stream -}
-recvLine (Connection buf h) n =
-  modifyMVar buf $ \bs ->
-    go (B.length bs) bs []
+recvLine (Connection bufVar hVar) n =
+  modifyMVar bufVar $ \bs ->
+    do h <- readMVar hVar
+       go h (B.length bs) bs []
   where
     -- bsn: cached length of concatenation of (bs:bss)
     -- bs : most recent chunk
     -- bss: other chunks ordered from most to least recent
-    go bsn bs bss =
+    go h bsn bs bss =
       case B8.elemIndex '\n' bs of
         Just i -> return (B.tail b, -- tail drops newline
                           Just (cleanEnd (B.concat (reverse (a:bss)))))
@@ -491,7 +491,7 @@ recvLine (Connection buf h) n =
              if B.null more -- connection closed
                then if bsn == 0 then return (B.empty, Nothing)
                                 else throwIO LineTruncated
-               else go (bsn + B.length more) more (bs:bss)
+               else go h (bsn + B.length more) more (bs:bss)
 
 
 -- | Push a 'ByteString' onto the buffer so that it will be the first
@@ -502,8 +502,8 @@ putBuf ::
   Connection {- ^ connection         -} ->
   ByteString {- ^ new head of buffer -} ->
   IO ()
-putBuf (Connection buf h) bs =
-  modifyMVar_ buf (\old -> return $! B.append bs old)
+putBuf (Connection bufVar _) bs =
+  modifyMVar_ bufVar (\old -> return $! B.append bs old)
 
 
 -- | Remove the trailing @'\\r'@ if one is found.
@@ -521,9 +521,25 @@ send ::
   Connection {- ^ open connection -} ->
   ByteString {- ^ chunk           -} ->
   IO ()
-send (Connection _ h) = networkSend h
+send (Connection _ hVar) bs =
+  do h <- readMVar hVar
+     networkSend h bs
 
 
+upgradeTls ::
+  TlsParams {- ^ connection params -} ->
+  String {- ^ hostname -} ->
+  Connection ->
+  IO ()
+upgradeTls tp hostname (Connection bufVar hVar) =
+  modifyMVar_ bufVar $ \buf ->
+  modifyMVar  hVar   $ \h ->
+  case h of
+    SSL{} -> return (h, buf)
+    Socket s ->
+      do (cert, ssl) <- startTls tp hostname (pure s)
+         return (SSL cert ssl, B.empty)
+
 ------------------------------------------------------------------------
 
 
@@ -542,34 +558,54 @@ startTls tp hostname mkSocket = SSL.withOpenSSL $
 
      -- configure context
      SSL.contextSetCiphers          ctx (tpCipherSuite tp)
-     installVerification            ctx hostname
-     SSL.contextSetVerificationMode ctx (verificationMode (tpInsecure tp))
+     traverse_ (contextSetTls13Ciphers ctx) (tpCipherSuiteTls13 tp)
+     case tpVerify tp of
+       VerifyDefault ->
+         do installVerification ctx hostname
+            SSL.contextSetVerificationMode ctx verifyPeer
+       VerifyHostname h ->
+         do installVerification ctx h
+            SSL.contextSetVerificationMode ctx verifyPeer
+       VerifyNone    -> pure ()
      SSL.contextAddOption           ctx SSL.SSL_OP_ALL
      SSL.contextRemoveOption        ctx SSL.SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS
 
+
      -- configure certificates
      setupCaCertificates ctx (tpServerCertificate tp)
      clientCert <- traverse (setupCertificate ctx) (tpClientCertificate tp)
-     traverse_ (setupPrivateKey ctx (tpClientPrivateKeyPassword tp)) (tpClientPrivateKey tp)
+
+     for_ (tpClientPrivateKey tp) $ \path ->
+       withDefaultPassword ctx (tpClientPrivateKeyPassword tp) $
+         SSL.contextSetPrivateKeyFile ctx path
 
      -- add socket to context
      -- creation of the socket is delayed until this point to avoid
      -- leaking the file descriptor in the cases of exceptions above.
      ssl <- SSL.connection ctx =<< mkSocket
 
-     -- configure hostname used for certificate validation
-     SSL.setTlsextHostName ssl hostname
+     -- configure hostname used for SNI
+     isip <- isIpAddress hostname
+     unless isip (SSL.setTlsextHostName ssl hostname)
 
      SSL.connect ssl
 
      return (clientCert, ssl)
 
+isIpAddress :: HostName -> IO Bool
+isIpAddress host =
+ do res <- try (Socket.getAddrInfo
+                  (Just Socket.defaultHints{Socket.addrFlags=[Socket.AI_NUMERICHOST]})
+                  (Just host) Nothing)
+    case res :: Either IOError [AddrInfo] of
+      Right{} -> pure True
+      Left {} -> pure False
 
 setupCaCertificates :: SSLContext -> Maybe FilePath -> IO ()
 setupCaCertificates ctx mbPath =
   case mbPath of
     Nothing   -> contextLoadSystemCerts ctx
-    Just path -> SSL.contextSetCAFile ctx path
+    Just path -> withDefaultPassword ctx Nothing (SSL.contextSetCAFile ctx path)
 
 
 setupCertificate :: SSLContext -> FilePath -> IO X509
@@ -578,36 +614,28 @@ setupCertificate ctx path =
      SSL.contextSetCertificate ctx x509
      pure x509
 
-
-setupPrivateKey :: SSLContext -> PEM.PemPasswordSupply -> FilePath -> IO ()
-setupPrivateKey ctx password path =
-  do str <- readFile path -- EX
-     key <- PEM.readPrivateKey str password
-     SSL.contextSetPrivateKey ctx key
-
-
-verificationMode :: Bool {- ^ insecure -} -> SSL.VerificationMode
-verificationMode insecure
-  | insecure  = SSL.VerifyNone
-  | otherwise = SSL.VerifyPeer
-                  { SSL.vpFailIfNoPeerCert = True
-                  , SSL.vpClientOnce       = True
-                  , SSL.vpCallback         = Nothing
-                  }
+verifyPeer :: SSL.VerificationMode
+verifyPeer = SSL.VerifyPeer
+  { SSL.vpFailIfNoPeerCert = True
+  , SSL.vpClientOnce       = True
+  , SSL.vpCallback         = Nothing
+  }
 
 -- | Get peer certificate if one exists.
 getPeerCertificate :: Connection -> IO (Maybe X509.X509)
-getPeerCertificate (Connection _ h) =
+getPeerCertificate (Connection _ hVar) =
+  withMVar hVar $ \h ->
   case h of
     Socket{} -> return Nothing
     SSL _ ssl -> SSL.getPeerCertificate ssl
 
 -- | Get peer certificate if one exists.
-getClientCertificate :: Connection -> Maybe X509.X509
-getClientCertificate (Connection _ h) =
-  case h of
-    Socket{} -> Nothing
-    SSL c _  -> c
+getClientCertificate :: Connection -> IO (Maybe X509.X509)
+getClientCertificate (Connection _ hVar) =
+  do h <- readMVar hVar
+     return $ case h of
+                Socket{} -> Nothing
+                SSL c _  -> c
 
 getPeerCertFingerprintSha1 :: Connection -> IO (Maybe ByteString)
 getPeerCertFingerprintSha1 = getPeerCertFingerprint "sha1"
